#############################################################################
# celltimes_plugin.tcl Copyright 2005 r0lZ
# PgcEdit plugin to convert CellTimes.txt to timecodes and vice-versa.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################
# Usage: copy this file in the <PgcEdit's install dir>/plugins directory.
# A "Celltimes" menu will be added in the Plugins menu.
# You need PgcEdit 0.6.0 or more to use this plugin.
#############################################################################
# Version History:
#
# v 1.0 (September 12, 2005)
# - First public release
#############################################################################


# Add Plugins -> Celltimes submenu
if {! [winfo exists .mainmenu.plugins.celltimes]} {
	.mainmenu.plugins add cascade -label "Cell-Zeiten" -menu .mainmenu.plugins.celltimes
	menu .mainmenu.plugins.celltimes
	# add Search & Replace with GUI item
	.mainmenu.plugins.celltimes add command -label "Konvertiere Frames zu Zeitcodes" -command \
		{::plugins::celltimes::frames2codes}
	.mainmenu.plugins.celltimes add command -label "Konvertiere Zeitcodes zu Frames" -command \
		{::plugins::celltimes::codes2frames}
	.mainmenu.plugins.celltimes add separator
	.mainmenu.plugins.celltimes add command -label "ber..." -command {
			tk_messageBox  -title "Plugin Cell-Zeiten" -icon info \
					-message "Plugin Cell-Zeiten fr PgcEdit 0.6.0 oder hher.\nVersion $::plugins::celltimes::version\nAutor: r0lZ\nFreeware!"
		}
}

namespace eval ::plugins {}
namespace eval ::plugins::celltimes {

set version "1.0"

#### GUI

proc input_args {} {
	if {! [info exists ::plugins::celltimes::rate]} {
		set ::plugins::celltimes::rate 25
		set ::plugins::celltimes::initialfile [file nativename [file join \
				[file dirname $::config(dumpinitialfile)] "Cell-Zeiten.txt"]]
	}
	set w .celltimes
	toplevel $w -class Dialog
	wm withdraw $w
	wm title $w "Cell-Zeiten"
	if {$::config(tk) != "aqua"} {wm transient $w .}
	if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
	wm protocol $w WM_DELETE_WINDOW {set ::plugins::celltimes::celltimes_done 0}
	bind $w <Escape> "set ::plugins::celltimes::celltimes_done 0"
	bind $w <Return> "set ::plugins::celltimes::celltimes_done 1"

	frame $w.f -bd 2 -relief groove
	pack  $w.f -padx 4 -pady 3 -fill x
	label $w.f.srcl -text "Auswhlen der Bildwiederholrate:"
	pack  $w.f.srcl -anchor w
	radiobutton $w.f.pal  -text "PAL:  25 FPS."  -variable ::plugins::celltimes::rate -value 25
	pack  $w.f.pal  -anchor w
	radiobutton $w.f.ntsc -text "NTSC: 30 FPS." -variable ::plugins::celltimes::rate -value 30
	pack  $w.f.ntsc -anchor w
	radiobutton $w.f.film -text "FILM: 24 FPS." -variable ::plugins::celltimes::rate -value 24
	pack  $w.f.film -anchor w

	# buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 4 -fill x
	frame  $w.fb.pad -bd 2 -relief groove
	pack   $w.fb.pad -fill x -expand 1 -side left
	button $w.fb.ok -text "OK" -width 8 -command {set ::plugins::celltimes::celltimes_done 1}
	pack   $w.fb.ok -side left -padx 2
	button $w.fb.cancel -text "Abbrechen" -width 8 -command {set ::plugins::celltimes::celltimes_done 0}
	pack   $w.fb.cancel -side left -padx 2

	update
	wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
	wm maxsize $w [winfo reqwidth $w] [winfo reqheight $w]
	set wx [expr {[winfo rootx .]+(([winfo width  .]-[winfo reqwidth  $w])/2)}]
	set wy [expr {[winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25}]
	if {$wx < 10} {set wx 10}
	if {$wy < 22} {set wy 22}
	wm geometry $w +$wx+$wy
	wm deiconify $w
	focus $w
	grab set $w

	tkwait variable ::plugins::celltimes::celltimes_done
	destroy $w
	if {! $::plugins::celltimes::celltimes_done} {return ""}

	set infile [tk_getOpenFile \
		-filetypes {{"Text Dateien" {*.txt}} {"Alle Dateien" {*.*}}} \
		-initialdir "[file dirname $::plugins::celltimes::initialfile]" \
		-initialfile "[file tail $::plugins::celltimes::initialfile]" \
		-title "ffnen der Quelldatei mit den Cell-Zeiten" -parent .]
	if {$infile == ""} {return ""}
	set ::plugins::celltimes::initialfile $infile

	set indata [readFile $infile auto]
	if {$indata == ""} {return ""}
	set indata [string map {" " "" "\t" ""} [string trim $indata \n]]
	set indata [split $indata \n]
	set emptylines [lsearch -all $indata {}]
	for {set i [expr [llength $emptylines]-1]} {$i >= 0} {incr i -1} {
		set indata [lreplace $indata [lindex $emptylines $i] [lindex $emptylines $i]]
	}
	return $indata
}

proc frames2codes {} {
	set flist [input_args]
	if {$flist == ""} return
	set fps $::plugins::celltimes::rate
	set out ""
	set errors ""
	set lasttc -1
	foreach f $flist {
		if {![string is integer $f]} {
			set errors "${errors}String \"$f\" ist kein Integer!  Zeitcode bersprungen!\n"
			continue
		}
		if {$f < 0} {
			set errors "${errors}String \"$f\" ist kein negativer Integer!  Zeitcode bersprungen!\n"
			continue
		}
		if {$f < $lasttc} {
			set errors "${errors}Warnung: Frameanzahl $f ist niedriger als $lasttc!\n"
		}
		set ff [expr $f % $fps]
		set ss [expr ($f / $fps) % 60]
		set mm [expr ($f / ($fps*60)) % 60]
		set hh [expr ($f / ($fps*3600))]
		set out ${out}[format "%02d:%02d:%02d.%02d\n" $hh $mm $ss $ff]
		set lasttc $f
	}
	if {$out == ""} {
		set errors "$errors\nAusgabe ist leer!"
	}
	if {$errors != ""} {
		if {[string length $errors] < 1000} {
			tk_messageBox -message "Fehler entdeckt:\n$errors" -icon error -title "Cell-Zeiten"
		} else {
			tk_messageBox -message "Fehler entdeckt.\nSiehe Protokoll fr Details." \
					-icon error -title "Cell-Zeiten"
			::main::showtext .celltimeserr "Cell-Zeiten Protokoll" $errors
		}
	}
	if {$out != ""} {
		::main::showtext .celltimesout "Cell-Zeiten als Zeitcodes" $out
	}
}

proc codes2frames {} {
	set tclist [input_args]
	if {$tclist == ""} return
	set fps $::plugins::celltimes::rate
	set out ""
	set errors ""
	set lasttc -1
	foreach tc $tclist {
		set dotpos [string first "." $tc]
		if {$dotpos > -1} {
			set ff [string range $tc [expr $dotpos+1] end]
			set tc2 [string range $tc 0 [expr $dotpos-1]]
			set n [scan $tc2 "%d:%d:%d" v(1) v(2) v(3)]
			incr n
			set v($n) $ff
		} else {
			set n [scan $tc "%d:%d:%d:%d" v(1) v(2) v(3) v(4)]
		}
		set offset [expr 4-$n]
		for {set i 1} {$i <= $n} {incr i} {
			set ii [expr $i+$offset]
			set vv($ii) [string trimleft $v($i) 0]
			if {$vv($ii) == ""} {set vv($ii) 0}
		}
		for {set i 1} {$i <= $offset} {incr i} {
			set vv($i) 0
		}
		set e false
		for {set i 1} {$i <= 4} {incr i} {
			if {! [string is integer $vv($i)]} {
				set errors "${errors}Fehler im Zeitcode \"$tc\" : String \"$vv($i)\" ist kein Integer!  Zeitcode bersprungen!\n"
				set e true
				break
			}
			if {$vv($i)<0} {
				set errors "${errors}Fehler im Zeitcode \"$tc\" : \"$vv($i)\" ist kein negativer Integer!  Zeitcode bersprungen!\n"
				set e true
				break
			}
		}
		if {$e} continue
		if {$vv(4) >= $fps} {
			set errors "${errors}Fehler im Zeitcode \"$tc\" : $vv(4) Frames sind >= $fps FPS!  Zeitcode bersprungen!\n"
			continue
		}
		if {$vv(3) >= 60} {
			set errors "${errors}Fehler im Zeitcode \"$tc\" : $vv(3) Sekunden sind >= 60!  Zeitcode bersprungen!\n"
			continue
		}
		if {$vv(2) >= 60} {
			set errors "${errors}Fehler im Zeitcode \"$tc\" : $vv(2) minutes >= 60!  Zeitcode bersprungen!\n"
			continue
		}
		set f [expr ($vv(1)*3600*$fps) + ($vv(2)*60*$fps) + ($vv(3)*$fps) + $vv(4)]
		set out "${out}$f\n"
		if {$f <= $lasttc} {
			set errors "${errors}Warnung im Zeitcode \"$tc\" : Frameanzahl $f ist niedriger als Frame $lasttc!\n"
		}
		set lasttc $f
	}
	if {$out == ""} {
		set errors "$errors\nAusgabe ist leer!"
	}
	if {$errors != ""} {
		if {[string length $errors] < 1000} {
			tk_messageBox -message "Fehler entdeckt:\n$errors" -icon error -title "Cell-Zeiten"
		} else {
			tk_messageBox -message "Fehler entdeckt.\nSiehe Protokoll fr Details." \
					-icon error -title "Cell-Zeiten"
			::main::showtext .celltimeserr "Cell-Zeiten Protokoll" $errors
		}
	}
	if {$out != ""} {
		::main::showtext .celltimesout "Cell-Zeiten als Frames" $out
	}
}

}	;# end of namespace ::plugins::celltimes
