#############################################################################
# hexbitoperands_plugin.tcl   Copyright 2008, 2009 Morris/Elysium
# Allows display of bit operations values in hex
# Allows display descriptive bit operands
#
# Portions copyright 2004-2008 r0lZ
#
# This plugin is freeware.
# Any comments are welcome. Send them to morris@elysium.pl
# 
# Version 0.7, 2008.12.29
# - initial release
#
#
# Version 0.8, 2009.02.05
# - added option to display hexadecimal values for all Set operations
# - added option to display descriptions for IF operations
#############################################################################


namespace eval ::plugins {}

namespace eval ::plugins::hexbitoperands {

proc set_range_values {rangestart rangeend rangestr rangecount multirange} {
  upvar 1 $rangestr strval
  upvar 1 $rangecount cntval
  upvar 1 $multirange multi
  switch [expr {$rangeend - $rangestart}] {
    0 {
      set strval "$strval, $rangestart"
    }

    1 {
      set strval "$strval, $rangestart, $rangeend"
      incr cntval
    }

    default {
      set strval "$strval, $rangestart\-$rangeend"
      set multi true
    }
  }
  incr cntval
}

proc set_bits_string {prefix_string rangestr rangecount multirange} {
  if {$rangecount == 1 && !$multirange} {
    set txt "Bit $rangestr"
  } else {
    set txt "Bits $rangestr"
  }

  return "$prefix_string $txt"
}

#############################################################################
## Return a word (2 bytes) of the hex command in decimal form.
## The argument 'bytenum' is the index of the first byte of the word (0-6)
## If 'showascii' is true, then append the ascii value of the word.
## The arguments 'operator and 'ifoperator' are needed to determine whether
## to use or not hexadecimal output of values

proc get_word_plugin {hex bytenum showascii {operator -1} {ifoperator -1}} {
  set r [lindex $hex $bytenum][lindex $hex [expr {$bytenum + 1}]]

  set bit_operation_description false
  set use_hex false

  set normalopidx [lsearch {1 3 4 5 6 7 8} $operator]
  set bitopidx [lsearch {9 10 11} $operator]

  if {$normalopidx != -1 || $bitopidx != -1 || $ifoperator == 1} {
    if {$normalopidx != -1} {
      set use_hex $::config(hexalloperands)
    } else {
      set use_hex $::config(hexbitoperands)
    }

    if {$bitopidx != -1} {
      set bit_operation_description $::config(descriptivebitoperands)
    } elseif {$ifoperator == 1} {
      set bit_operation_description $::config(descriptivebitifoperands)
    }
  }

  if {$use_hex} {
    set r "0x[string toupper $r]"
  } else {
    set r [x2d $r]
  }

  if {$showascii} {
    if {!$bit_operation_description} {
      set ascii ""
      set num $r
      while {$num > 0} {
        set c [expr {$num % 256}]
        set num [expr {$num / 256}]
        if {$c <= 0X1F || ($c >= 0X80 && $c <= 0X9F)} {
          set ascii ".$ascii"
        } else {
          set ascii "[format {%c} $c]$ascii"
        }
      }
      while {[string length $ascii] < 2} {
        set ascii ".$ascii"
      }
      if {[::streams::valid_langcode $ascii]} {
        set r "$r (\"${ascii}\")"
      }
    } else {
      set num $r
      set bitnum 0
      set on_range_start -1
      set on_range_end -1
      set off_range_start -1
      set off_range_end -1
      set on_bitnums ""
      set off_bitnums ""
      set on_range_count 0
      set off_range_count 0
      set on_multirange_exists false
      set off_multirange_exists false
      
      while {$bitnum < 16} {
        if {[expr {$num & 1}] == 1} {

          if {$off_range_start >= 0} {
            ::plugins::hexbitoperands::set_range_values $off_range_start $off_range_end off_bitnums off_range_count off_multirange_exists
            set off_range_start -1
          }

          if {$on_range_start == -1} {
            set on_range_start $bitnum
          }
          set on_range_end $bitnum

        } else {

          if {$on_range_start >= 0} {
            ::plugins::hexbitoperands::set_range_values $on_range_start $on_range_end on_bitnums on_range_count on_multirange_exists
            set on_range_start -1
          }

          if {$off_range_start == -1} {
            set off_range_start $bitnum
          }
          set off_range_end $bitnum

        }
        set num [expr {$num >> 1}]
        incr bitnum
      }

      if {$on_range_start >= 0} {        
        ::plugins::hexbitoperands::set_range_values $on_range_start $on_range_end on_bitnums on_range_count on_multirange_exists
      }

      if {$off_range_start >= 0} {
        ::plugins::hexbitoperands::set_range_values $off_range_start $off_range_end off_bitnums off_range_count off_multirange_exists
      }

      # delete ", " from the string beginning
      if {[string index $on_bitnums 0] == ","} {
        set on_bitnums [string replace $on_bitnums 0 1]
      }

      if {[string index $off_bitnums 0] == ","} {
        set off_bitnums [string replace $off_bitnums 0 1]
      }

      set description ""
      set num $r
      if {$ifoperator == 1} {
            if {$num != 0 && $num != 0xFFFF} {
              set description [::plugins::hexbitoperands::set_bits_string "berprfen fr" $on_bitnums $on_range_count $on_multirange_exists]
            }
      } else {
        switch $operator {
          9 { ;# AND
            if {$num == 0} {
              set description "lscht alle Bits"
            } elseif {$num == 0xFFFF} {
              set description "behlt alle Bits"
            } else {
              if {$on_range_count > $off_range_count} {
                set description [::plugins::hexbitoperands::set_bits_string "lscht" $off_bitnums $off_range_count $off_multirange_exists]
              } else {
                set description [::plugins::hexbitoperands::set_bits_string "behlt" $on_bitnums $on_range_count $on_multirange_exists]
              }
            }

          }

          10 { ;# OR
            if {$num != 0} {
              if {$num == 0xFFFF} {
                set description "setzt alle Bits"
              } else {
                set description [::plugins::hexbitoperands::set_bits_string "setzt" $on_bitnums $on_range_count $on_multirange_exists]
              }
            }
          }

          11 { ;# XOR
            if {$num != 0} {
              if {$num == 0xFFFF} {
                set description "ndert alle Bits"
              } else {
                set description [::plugins::hexbitoperands::set_bits_string "ndert" $on_bitnums $on_range_count $on_multirange_exists]
              }
            }
          }
        }
      }
      if {$description != ""} {set r "$r ($description)"}
    }
  }
  return $r
}
#############################################################################

proc load_config {} {
  set cfgfile [file join $::config(config_directory) "hexbitoperands.cfg"]
  if {[file exists $cfgfile]} {
    catch {source $cfgfile}
  }
}

proc update_config {} {
  set config "set ::config(hexbitoperands) $::config(hexbitoperands)"
  set config "$config\nset ::config(hexalloperands) $::config(hexalloperands)"
  set config "$config\nset ::config(descriptivebitoperands) $::config(descriptivebitoperands)"
  set config "$config\nset ::config(descriptivebitifoperands) $::config(descriptivebitifoperands)"
  set config "$config\n"
  catch {saveFile [file join $::config(config_directory) "hexbitoperands.cfg"] auto $config}
}

proc option_updated {} {
  if {$::config(hexbitoperands) || $::config(hexalloperands) || $::config(descriptivebitoperands) || $::config(descriptivebitifoperands)} {

    if {!$::plugins::hexbitoperands::substitute_enabled} {

      # Modified by r0lZ: check compatibility with PgcEdit version
      if {![test_version 9.1 "Hex-Bit-Operanden Plugin"]} {

        set ::config(hexbitoperands) false
        set ::config(hexalloperands) false
        set ::config(descriptivebitoperands) false
        set ::config(descriptivebitifoperands) false

    } else {

        set ::plugins::hexbitoperands::substitute_enabled true

        # substitute original "get_word" proc with the one provided by the plugin
        rename ::disasm::get_word                         ::disasm::get_word_pgcedit
        rename ::plugins::hexbitoperands::get_word_plugin ::disasm::get_word
      }
    }

  } else {

    if {$::plugins::hexbitoperands::substitute_enabled} {

      set ::plugins::hexbitoperands::substitute_enabled false

      # restore original "get_word" proc 
      rename ::disasm::get_word         ::plugins::hexbitoperands::get_word_plugin
      rename ::disasm::get_word_pgcedit ::disasm::get_word
    }

  }
  ::main::init_refresh_listbox
  ::main::refresh_listbox

  # moved here by r0lZ
  ::plugins::hexbitoperands::update_config
}

proc about_plugin {} {
  set txt "Hex-Bit-Operanden Plugin fr PgcEdit 9.1 oder hher.\n"
  set txt "${txt}Version 0.8, 05.02.2009\n"
  set txt "${txt}Geschrieben von Morris/Elysium\n\n"
  set txt "${txt}Dieses Plugin ist Freeware!"
  tk_messageBox -title "Hex-Bit-Operanden Plugin" -icon info -message $txt
}

} ;# end of ::plugins::hex


# Init plugin

set ::config(hexbitoperands) true
set ::config(hexalloperands) false
set ::config(descriptivebitoperands) true
set ::config(descriptivebitifoperands) false

set ::plugins::hexbitoperands::substitute_enabled false

# try to load config from file
::plugins::hexbitoperands::load_config

# force refresh of the main window (because the original disasm proc has already been executed)
::plugins::hexbitoperands::option_updated

.mainmenu.plugins add cascade -label "Hex-Bit-Operanden" -menu .mainmenu.plugins.hexbitoperands
  menu .mainmenu.plugins.hexbitoperands
  
  .mainmenu.plugins.hexbitoperands add checkbutton -label "Hexadezimal-Werte fr Bit-Operationen anzeigen" \
    -variable ::config(hexbitoperands) -onvalue true -offvalue false -command {::plugins::hexbitoperands::option_updated}

  .mainmenu.plugins.hexbitoperands add checkbutton -label "Hexadezimal-Werte fr keine Bit-Operationen anzeigen" \
    -variable ::config(hexalloperands) -onvalue true -offvalue false -command {::plugins::hexbitoperands::option_updated}

  .mainmenu.plugins.hexbitoperands add checkbutton -label "Beschreibung fr Bit-Operanden anzeigen" \
    -variable ::config(descriptivebitoperands) -onvalue true -offvalue false -command {::plugins::hexbitoperands::option_updated}

  .mainmenu.plugins.hexbitoperands add checkbutton -label "Beschreibung der Bit-Operanden fr den Vergleich anzeigen" \
    -variable ::config(descriptivebitifoperands) -onvalue true -offvalue false -command {::plugins::hexbitoperands::option_updated}

  .mainmenu.plugins.hexbitoperands add separator
  .mainmenu.plugins.hexbitoperands add command -label "ber..." -command {::plugins::hexbitoperands::about_plugin}
