#############################################################################
# playall_plugin.tcl Copyright 2007, 2008, 2009 r0lZ
# PgcEdit plugin to create a new PGC with all cells in the VOB file(s).
#
# This plugin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This plugin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################
# Usage: copy this file in the <PgcEdit's install dir>/plugins directory.
# An "Play All PGC" menu will be added in the Plugins menu.
# You need PgcEdit 0.6.1 or more to use this plugin.
#############################################################################
# Version History:
#
# v 1.5 (December 6, 2008)
# - Requires PgcEdit v9.0 or greater.
# - Added the "Please wait" dialog with an Abort button when scanning the VOBs.
# v 1.4 (March 18, 2008)
# - Minor bug fix: After creating a PlayAll PGC in a menu domain, the menu was
#   flagged as having several non-identical LUs, even if the menu contains only
#   one LU. It was necessary to reload the DVD to be able to use some functions.
# v 1.3 (January 1, 2008)
# - Bug fix: Error when creating a play all PGC from the VMGM domain.
# - Bug fix: The check for the minimal version of PgcEdit was wrong, and the
#   plugin was unable to work when the DVDShrink plugin was not installed.
# v 1.2 (May 15, 2007)
# - Bug fix: The last scanned VOB file was not closed after the process.  It
#   was necessary to close PgcEdit to release the lock.
# v 1.1 (May 2, 2007)
# - Little bug fixed: Calling Find unreferenced VOB cells in current domain
#   with the FP-PGC selected produced an error.
# v 1.0 (April 19, 2007)
# - First public release.  Requires PgcEdit v8.1+
#############################################################################


# Add Plugins -> Play All PGC submenu
if {! [winfo exists .mainmenu.plugins.playall]} {
	.mainmenu.plugins add cascade -label "Play All PGC" -menu .mainmenu.plugins.playall
	menu .mainmenu.plugins.playall
	.mainmenu.plugins.playall add command -label "Erstelle ein neues 'Play All' PGC von den VOBs" -command {::plugins::playall::play_all_pgc_from_vobs}
	.mainmenu.plugins.playall add command -label "Erstelle ein neues PGC, nur mit den nicht referenzierten VOB-Cells" -command {::plugins::playall::play_all_pgc_from_vobs true}
	.mainmenu.plugins.playall add separator
	.mainmenu.plugins.playall add command -label "Finde nicht referenzierte VOB-Cells in der aktuellen Domain" -command {::plugins::playall::find_unref_vob_cells_domain}
	.mainmenu.plugins.playall add command -label "Finde nicht referenzierte VOB-Cells in der ganzen DVD" -command {::plugins::playall::find_unref_vob_cells_dvd}
	.mainmenu.plugins.playall add separator
	.mainmenu.plugins.playall add command -label "ber..." -command {
			tk_messageBox  -title "Play All PGC - Plugin" -icon info \
					-message "Play All PGC - Plugin fr PgcEdit 9.0 oder hher.\nVersion $::plugins::playall::version\nAutor: r0lZ\nFreeware!"
		}
}


namespace eval ::plugins {}
namespace eval ::plugins::playall {

set version "1.5"

# Scan the VOB files for cells, and create a new PGC with all cells of the VOBs,
# or with only the cells that are not referenced in the IFO PGCs.
# TODO: better handling of the seamless joint flag
# TODO: better handling of interleaved cells
proc play_all_pgc_from_vobs {{unrefonly false}} {
	if {! [test_version 8.7 "Play all plugin"]} return
	if {$unrefonly} {
		set toolname "Erstelle ein neues PGC mit nicht referenzierten VOB-Cells"
	} else {
		set toolname "Erstelle ein neues 'Play All' PGC von den VOBs"
	}
	if {![test_version 8.1 "Play All PGC Plugin"]} return
	if {![info exists ::dvddir]} {bell; return}
	set type [lindex $::editor_cmds_from 0]
	set vts [lindex [split [lindex $::editor_cmds_from 1] ,] 0]
	if {$type == "pgc"} {
		if {$vts == 0} {bell; return}
		set newpgc [expr $::pgcs($vts,numpgcs)+1]
		set vobfile [file join $::dvddir [format "VTS_%02d_1.VOB" $vts]]
		set vobsize 0
		for {set i 1} {$i < 10} {incr i} {
			set vobfilename [file join $::dvddir VTS_[format %02i $vts]_$i.VOB]
			if {[file exists $vobfilename]} {
				incr vobsize [expr [file size $vobfilename] / 2048]
			}
		}
		set orig_pgcheader $::pgcs([lindex $::editor_cmds_from 1],header)
	} else {
		set lu [lindex [split [lindex $::editor_cmds_from 1] ,] 1]
		set newpgc [expr $::menus($vts,$lu,numpgcs)+1]
		if {$vts == 0} {
			set vobfile [file join $::dvddir "VIDEO_TS.VOB"]
		} else {
			set vobfile [file join $::dvddir [format "VTS_%02d_0.VOB" $vts]]
		}
		set vobsize [expr {[file size $vobfile]/2048}]
		set orig_pgcheader $::menus([lindex $::editor_cmds_from 1],header)
	}
	set_int orig_pgcheader 0 0 4
	set_int orig_pgcheader 4 0 4
	set_int orig_pgcheader 8 0 4
	set_int orig_pgcheader 0x9C 0 4
	set_int orig_pgcheader 0xA0 0 4

	if {! [file exists $vobfile]} {
		set txt "Kann die VOB-Datei [file tail $vobfile] nicht finden!"
		set txt "$txt\nDa dieses Werkzeug die VOB-Datei(en) der aktuellen Domain analysiert,"
		set txt "$txt\nzum Erstellen des neuen 'Play All' PGC mit allen Cells in der VOB,"
		set txt "$txt\nmuss die VOB-Datei bzw. die VOB-Dateien im DVD-Ordner vorhanden sein."
		tk_messageBox -message $txt -title $toolname -icon error
		return
	}

	set txt "Diese Funktion wird ein neues PGC am Ende der aktuellen Domain erstellen und enthlt"
	if {$unrefonly} {
		set txt "$txt\nalle vorhandenen Cells in den VOB-Dateien der Domain, aber keine Referenzen in den IFOs."
	} else {
		set txt "$txt\nalle vorhandenen Cells in den VOB-Dateien von der aktuellen Domain."
	}
	set txt "$txt\n\nHINWEISE:"
	set txt "$txt\n- Das Seamless Joint Flag wird nur gesetzt, wenn sich die VOB-ID nicht ndert !"
	set txt "$txt\n- Alle Interleaved Cells sind in den Cell-Tabellen einer Multi-PGC Story miteinbezogen !"
	set txt "$txt\n- Die Audio- und Subpicture-Streams werden vom aktuellen PGC kopiert."
	set txt "$txt\n- Die Farbzuordnungstabelle (Clut) der Subpictures werden auch vom aktuellen PGC kopiert."
	set txt "$txt\n- Die Navigationsparameter und die VM-Commands werden nicht erstellt."
	set txt "$txt\n- In den Men Domains, wird das neue PGC nur in der aktuellen LU erstellt."
	set txt "$txt\n- Sie werden gewarnt, wenn es zu viele Cells, Programme oder Kapitel im PGC gibt."
	set txt "$txt\n- Diese Funktion kann eine lange Zeit in Anspruch nehmen.  Bitte haben Sie Geduld."
	set txt "$txt\n\nFortfahren ?"
	set r [tk_messageBox -message $txt -title $toolname -icon info -type yesno -default yes \
			-skip "playall_plugin_confirm1"]
	if {$r == "no"} return

	set ::blocking_msg_aborted false
	set blocking [blocking_msg "Suchen nach Cells in den VOB-Dateien." true]

	# determine frame rate
	if {$vts == 0} {
		if {$::pgcs(numvts) >= 1} {
			set frvts 1
			set frpgc 1
		} else {
			set frvts 0
			set frpgc 0
		}
	} else {
		set frvts $vts
		set frpgc 1
	}
	if {[expr [get_int ::pgcs($frvts,$frpgc,header) 7 1] & 0xC0] == 192} {
		set fps 30
	} else {
		set fps 25
	}

	if {$unrefonly} {
		set referencedcells [findReferencedCells $type $vts]
	}

	set multiangle false
	if {$type == "pgc"} {
		for {set p 1} {$p <= $::pgcs($vts,numpgcs)} {incr p} {
			set multiangle [findAngles $multiangle $::pgcs($vts,$p,cellpbtable)]
		}
	}

	set err false
	set old_vobfile ""
	set lba 0
	set sector 0
	set last_vid -1
	set last_cid -1
	set celltimeoffset 0
	set fh ""
	set progressbarlimit 0
	set resultlist {}
	while {1} {
		if {$vobfile != $old_vobfile} {
			if {![file exists $vobfile]} {
				set err true
				set errmsg "Datei nicht gefunden"
				break
			}
			if {$fh != ""} {
				close $fh
				set old_vobfile ""
			}
			set fh [::menued::openbinfile $vobfile r]
			if {$fh == ""} {
				set err true
				set errmsg "Kann nicht die Datei [file tail $vobfile] ffnen"
				break
			}
			set old_vobfile $vobfile
		}
		set navpack [::menued::load_nav_pack $vobfile $fh $sector]
		if {$navpack == ""} {
			set err true
			set errmsg "Nav-Pack nicht gltig oder LBA $lba auerhalb des Bereiches"
			break
		}
		if {$::blocking_msg_aborted} {
			break
		}
		set vobuend  [get_int navpack 0x40F 4]
		set vid      [get_int navpack 0x41F 2]
		set cid      [get_int navpack 0x422 1]

		if {! $unrefonly || [lsearch $referencedcells [list $vid $cid]] == -1} {
			if {! [info exists cells($vid,$cid,entryvobu)]} {
				# a new cell begins at this lba
				# keep track of the beginning of the new cell
				lappend resultlist [list $vid $cid]
				set cells($vid,$cid,entryvobu) $lba
				set cells($vid,$cid,ilved) false
				set cells($vid,$cid,timing) 0
				set cells($vid,$cid,firstvobustartptm) [get_int navpack 57 4]
				show_msg "Scanne $vobfile, VOB/CELL-ID $vid/$cid.  Bitte warten..." $lba $vobsize
				set progressbarlimit 0
			} elseif {$last_vid != $vid || $last_cid != $cid} {
				set cells($vid,$cid,ilved) true
			}
			if {($last_vid != $vid || $last_cid != $cid) && \
					![info exists cells($last_vid,$last_cid,firstilvuend)]} {
				set cells($last_vid,$last_cid,firstilvuend) [expr $lba-1]
			}
			set cells($vid,$cid,lastvobustart) $lba
			set cells($vid,$cid,lastvobuend) [expr $lba+$vobuend]
			set cells($vid,$cid,vobuendptm) [get_int navpack 61 4]
		}

		set last_vid $vid
		set last_cid $cid
		set lba [expr $lba + $vobuend + 1]
		set progressbarlimit [expr $progressbarlimit + $vobuend + 1]
		if {$progressbarlimit > 25000} {
			show_msg "Scanne $vobfile, VOB/CELL-ID $vid/$cid.  Bitte warten..." $lba $vobsize
			set progressbarlimit 0
		}
		set poslist [::menued::get_nav_pack_position $type $vts $lba $::dvddir]
		set vobfile [lindex $poslist 0]
		set sector  [lindex $poslist 1]

		if {$vobfile == ""} break
		if {$type == "pgc" && ! [file exists $vobfile]} break
		if {[expr [file size $vobfile] / 2048] < $sector} break
	}
	close $fh

	if {$::blocking_msg_aborted} {
		if {[catch {close_blocking_msg $blocked}]} {close_blocking_msg 1}
		show_msg ""
		return
	}


	# Build the PGC tables
	show_msg "Erstelle das neue 'Play All' PGC..." 100
	set numcells [llength $resultlist]
	if {$numcells == 0} {
		if {[catch {close_blocking_msg $blocked}]} {close_blocking_msg 1}
		show_msg "Keine nicht referenzierten Cells in den VOBs gefunden."
		set rtype $type
		if {$type == "pgc"} {set rtype title}
		set txt "Es gibt keine nicht referenzierten Cells in den VOB-Dateien von der $rtype Domain im VTS $vts."
		set txt "$txt\n\nVorgang abgebrochen."
		tk_messageBox -message $txt -title $toolname -icon info -skip playall_plugin_no_unref_cells
		return
	}

	set pgctime \000\000\000\000
	set_int pgctime 3 [expr [get_int ::pgcs($frvts,$frpgc,header) 7 1] & 0xC0] 1

	set pgmaptable {}
	set cellpbtable {}
	set cellpostable {}
	set last_vid -1
	set last_lastvobuend -999
	set angle 0
	set nextangle 0
	set maxangle 0
	set pg 1
	set last_ilved false
	set warning_seamless_flag false

	for {set cell 1} {$cell <= $numcells} {incr cell} {
		set c [lindex $resultlist [expr $cell-1]]
		set vid [lindex $c 0]
		set cid [lindex $c 1]
		set bcd [::tmap::Clock90KHz2BCD [expr $cells($vid,$cid,vobuendptm) - $cells($vid,$cid,firstvobustartptm)] $fps]
		set cellpbentry \000\000\000\000$bcd[string repeat \000 16]
		set cellflags 0
		if {$cells($vid,$cid,ilved)} {
			set angle $nextangle
			if {$cell < $numcells} {
				set next_c [lindex $resultlist $cell]
				set next_vid [lindex $next_c 0]
				set next_cid [lindex $next_c 1]
				set next_entryvobu $cells($next_vid,$next_cid,entryvobu)
				set next_bcd [::tmap::Clock90KHz2BCD [expr $cells($next_vid,$next_cid,vobuendptm) - $cells($next_vid,$next_cid,firstvobustartptm)] $fps]
			} else {
				set next_entryvobu [expr $cells($vid,$cid,lastvobuend)+1]
				set next_bcd ""
			}
			if {$multiangle} {
				incr angle
				if {$angle == 1} {
					if {$bcd == $next_bcd} {
						incr cellflags 85	;# first angle
					} else {
						set angle 0
					}
				} elseif {$cells($vid,$cid,lastvobuend) < $next_entryvobu} {
					incr cellflags 213	;# last angle
				} else {
					incr cellflags 149	;# middle angle
				}
				if {$angle > $maxangle} {
					set maxangle $angle
				}
				set nextangle $angle
				if {$cells($vid,$cid,lastvobuend) < $next_entryvobu} {
					set nextangle 0
				}
			} else {
				incr cellflags 4	;# interleaved allocation
			}
		} else {
			set angle 0
			set nextangle 0
		}
		if {($last_vid != $vid && $cid == 1) || \
				([expr $last_lastvobuend+1] != $cells($vid,$cid,entryvobu) && $cellflags == 0)} {
			incr cellflags 2
		}
		if {$cell > 1 && ($last_vid == $vid || $last_ilved || $cells($vid,$cid,ilved)) && \
			([expr $last_lastvobuend+1] == $cells($vid,$cid,entryvobu) || $cellflags > 2)} {
			incr cellflags 8
		} elseif {$cell > 1} {
			set warning_seamless_flag true
		}

		set_int cellpbentry 0 $cellflags 1
		set_int cellpbentry 8 $cells($vid,$cid,entryvobu) 4
		if {$cells($vid,$cid,ilved)} {
			set_int cellpbentry 12 $cells($vid,$cid,firstilvuend) 4
		}
		set_int cellpbentry 16 $cells($vid,$cid,lastvobustart) 4
		set_int cellpbentry 20 $cells($vid,$cid,lastvobuend) 4
		set cellpbtable $cellpbtable$cellpbentry

		set cellposentry \000\000\000\000
		set_int cellposentry 0 $vid 2
		set_int cellposentry 3 $cid 1
		set cellpostable $cellpostable$cellposentry
		if {$angle <= 1} {
			set pgctime [::pgced::AddBCDtime $pgctime $bcd]
			if {$pg < 100} {
				set pgmapentry \000
				set_int pgmapentry 0 $cell 1
				set pgmaptable $pgmaptable$pgmapentry
				incr pg
			}
		}

		set last_vid $vid
		set last_lastvobuend $cells($vid,$cid,lastvobuend)
		set last_ilved $cells($vid,$cid,ilved)
	}

	set pgcientry [string repeat \000 8]
	if {$type == "pgc"} {
		set ttn 0
		for {set p 1} {$p <= $::pgcs($vts,numpgcs)} {incr p} {
			if {$::pgcs($vts,$p,TTN) > $ttn} {
				set ttn $::pgcs($vts,$p,TTN)
			}
		}
		incr ttn
		set idx [expr {($newpgc * 8)}]
		set part1 [string range $::raw_sectors($vts,VTS_PGCI) 0 [expr {$idx-1}]]
		set part2 [string range $::raw_sectors($vts,VTS_PGCI) $idx end]
		set ::raw_sectors($vts,VTS_PGCI) $part1[string repeat \000 8]$part2
		set_int ::raw_sectors($vts,VTS_PGCI) $idx [expr {$ttn+128}] 1

		# update internal variables
		set ::pgcs($vts,$newpgc) {{* pre} {* post} {* cell}}
		set ::pgcs($vts,$newpgc,header) [string range $orig_pgcheader 0 3]$pgctime[string range $orig_pgcheader 8 end]
		set ::pgcs($vts,$newpgc,numprograms) [string length $pgmaptable]
		set ::pgcs($vts,$newpgc,programmap) $pgmaptable
		set ::pgcs($vts,$newpgc,numcells) $numcells
		set ::pgcs($vts,$newpgc,cellpbtable) $cellpbtable
		set ::pgcs($vts,$newpgc,cellpostable) $cellpostable
		set ::pgcs($vts,$newpgc,numbuttons) 0
		set ::ttns($vts,$ttn) $newpgc
		set ::pgcs($vts,$newpgc,TTN) $ttn

		set pttlist {}
		for {set p 0} {$p < $::pgcs($vts,$newpgc,numprograms)} {incr p} {
			set c [get_int pgmaptable $p 1]
			lappend pttlist [list [expr $p+1] [expr $p+1] $ttn]
		}
		set ::pgcs($vts,$newpgc,pttlist) $pttlist

		incr ::pgcs($vts,numpgcs)

		# Add the new title to the VMG global table of titles
		set data $::raw_sectors(0,VMG_TT_SRPT)
		set srptnumtitles [get_int data 0 2]
		set data [string range $data 0 [expr {($srptnumtitles*12)+7}]]
		incr srptnumtitles
		set_int data 0 $srptnumtitles 2

		set newentry \024\001\000\000\000\000\000\000\000\000\000\000
		if {$maxangle == 0} {set maxangle 1}
		set_int newentry 1 $maxangle 1
		set_int newentry 2 [llength $::pgcs($vts,$newpgc,pttlist)] 2
		set_int newentry 6 $vts 1
		set_int newentry 7 $::pgcs($vts,$newpgc,TTN) 1
		set data $data$newentry
		set_int data 4 [expr {[string length $data]-1}] 4
		set ::raw_sectors(0,VMG_TT_SRPT) $data
		set ::pgcs($vts,$newpgc,Title) $srptnumtitles
		set ::pgcs(Title,$srptnumtitles) [list $vts $newpgc]
		set ::pgcs(Title,NumTitles) $srptnumtitles

		# Add blank Time Map table in last position in VTS_TMAPT
		if {$::raw_sectors($vts,VTS_TMAPTI) != ""} {
			set numentries [get_int ::raw_sectors($vts,VTS_TMAPTI) 0 2]
			set table [string range $::raw_sectors($vts,VTS_TMAPTI) 0 7][string repeat \000 [expr 4*$::pgcs($vts,numpgcs)]]
			for {set i 1} {$i <= $numentries} {incr i} {
				set idx1 [get_int ::raw_sectors($vts,VTS_TMAPTI) [expr 4+(4*$i)] 4]
				set idx2 [expr $idx1+3+([get_int ::raw_sectors($vts,VTS_TMAPTI) [expr $idx1+2] 2]*4)]
				set tmap($i) [string range $::raw_sectors($vts,VTS_TMAPTI) $idx1 $idx2]
			}
			while {$i <= $::pgcs($vts,numpgcs)} {
				set tmap($i) \001\000\000\000
				incr i
			}
			set_int table 0 $::pgcs($vts,numpgcs) 2
			for {set i 1} {$i <= $::pgcs($vts,numpgcs)} {incr i} {
				set_int table [expr 4+(4*$i)] [string length $table] 4
				set table $table$tmap($i)
			}
			set ::raw_sectors($vts,VTS_TMAPTI) $table
			unset table
		}

		# create new PTT table
		set ptttable $::raw_sectors($vts,VTS_PTT_SRPT)
		set pttnumtitles [get_int ptttable 0 2]
		set pttheader [string range $ptttable 0 [expr {($pttnumtitles*4) + 7}]]
		set pttrest [string range $ptttable [expr {($pttnumtitles*4) + 8}] end]
		set newptts {}
		foreach i $::pgcs($vts,$newpgc,pttlist) {
			set pttentry \000\000\000\000
			set_int pttentry 0 [lindex $i 0] 2
			set_int pttentry 2 [lindex $i 2] 2
			set newptts $newptts$pttentry
		}
		set idx [string length $pttheader]
		set pttheader $pttheader\000\000\000\000
		set_int pttheader $idx [expr {[string length $pttheader]+[string length $pttrest]}] 4
		for {set i 0} {$i < $pttnumtitles} {incr i} {
			set idx [expr {($i * 4)+8}]
			set_int pttheader $idx [expr {[get_int pttheader $idx 4] +4}] 4
		}
		set ptttable $pttheader$pttrest$newptts
		set_int ptttable 0 $::pgcs($vts,$newpgc,TTN) 2
		set ::pgcs($vts,numttus) $::pgcs($vts,$newpgc,TTN)
		set_int ptttable 4 [expr {[string length $ptttable]-1}] 4
		set ::raw_sectors($vts,VTS_PTT_SRPT) $ptttable

		# update number of BOVs (if VTS already scanned)
		if {[info exists ::menued::scan_done] && $::menued::scan_done} {
			::menued::add_title_buttons_marker $vts
		}
		# create new time map table
		::main::init_tree_menu_no_move
		::main::display_pgc pgc  $vts $newpgc
		if {[catch {close_blocking_msg $blocked}]} {close_blocking_msg 1}
		::tmap::rebuild_tmap_from_vobs $vts $newpgc
		if {$::blocking_msg_aborted} {
			::tmap::need_new_tmap $vts $newpgc
		}
	} else {
		# Menus

		# insert new PGC at the end of the current LU
		if {$vts == 0} {
			set pgcitable $::raw_sectors(0,VMGM_PGCI_UT)
		} else {
			set pgcitable $::raw_sectors($vts,VTSM_PGCI_UT)
		}
		set luidx [get_int pgcitable [expr {($lu*8) + 4}] 4]
		set idx [expr {($::menus($vts,$lu,numpgcs) * 8) + $luidx + 8}]
		set part1 [string range $pgcitable 0 [expr {$idx-1}]]
		set part2 [string range $pgcitable $idx end]
		set pgcitable $part1[string repeat \000 8]$part2
		set_int pgcitable $idx 0 1
		if {$vts == 0} {
			set ::raw_sectors(0,VMGM_PGCI_UT) $pgcitable
		} else {
			set ::raw_sectors($vts,VTSM_PGCI_UT) $pgcitable
		}

		# update internal variables
		set ::menus($vts,$lu,$newpgc) {{* pre} {* post} {* cell}}
		set ::menus($vts,$lu,$newpgc,mask) 0
		set ::menus($vts,$lu,$newpgc,masktext) ""
		set ::menus($vts,$lu,$newpgc,header) [string range $orig_pgcheader 0 3]$pgctime[string range $orig_pgcheader 8 end]
		set ::menus($vts,$lu,$newpgc,numprograms) [string length $pgmaptable]
		set ::menus($vts,$lu,$newpgc,programmap) $pgmaptable
		set ::menus($vts,$lu,$newpgc,numcells) $numcells
		set ::menus($vts,$lu,$newpgc,cellpbtable) $cellpbtable
		set ::menus($vts,$lu,$newpgc,cellpostable) $cellpostable

		if {$::menus($vts,numlus) > 1} {
			set ::menus($vts,lus_are_identical) false
		}
		incr ::menus($vts,$lu,numpgcs)

		# update number of menu buttons
		::menued::add_menu_buttons_marker $vts
	}

	::main::init_tree_menu_no_move
	if {$type == "pgc"} {
		::main::display_pgc pgc  $vts $newpgc
	} else {
		::main::display_pgc menu $vts $lu $newpgc
	}
	::main::dvd_modified true false
	show_msg "Neues 'Play All' PGC erfolgreich erstellt."

	if {[catch {close_blocking_msg $blocked}]} {close_blocking_msg 1}

	if {$warning_seamless_flag} {
		set txt "WARNUNG: Das Seamless Joint Flag ist bei einigen Cells nicht gesetzt worden.\n"
		set txt "$txt\nMomentan versucht diese Funktion zu schtzen, ob das Seamless Flag gesetzt werden muss,"
		set txt "$txt\naber es kann hufig versagen!"
		set txt "$txt\n\nberprfen Sie bitte die Seamless Joint Flags im PGC-Editor."
		if {$type == "pgc"} {
			set txt "$txt\n\nSie knnen sie wahrscheinlich bei allen Cells desselben Titel setzen."
		}
		set txt "$txt\nEs ist fr gewhnlich sicher, das Flag an einer Cell zu setzen, die die vorherige fortsetzt."
		tk_messageBox -message $txt -title $toolname -icon warning -skip playall_plugin_warning_seamless_flag
	}

	# Warning if too many cells, PGs or PTTs
	if {$numcells > 99} {
		set txt "WARNUNG: Es gibt $numcells Cells in dem neuen 'Play All' PGC.\n"
		if {$type == "pgc"} {
			set txt "$txt\nDie maximale Anzahl von Programmen und Kapiteln ist 99."
			set txt "$txt\nNur 99 Programme und Kapitel sind erstellt worden."
		} else {
			set txt "$txt\nDie maximale Anzahl von Programmen ist 99."
			set txt "$txt\nNur 99 Programme sind erstellt worden."
		}
		if {$numcells > 255} {
			set txt "$txt\n\nAuerdem, die maximale Anzahl von Cells in einem PGC ist 255."
			set txt "$txt\nSie mssen mindestens [expr $numcells-255] Cells vom PGC manuell entfernen."
		}
		set txt "$txt\n\nDer PGC-Editor wird jetzt geffnet. Korrigieren Sie bitte die Cell-Tabelle!"
		tk_messageBox -message $txt -title $toolname -icon warning
		::main::action_Edit_PGC
	}
	return
}

proc addCells {referencedcells cellpostable} {
	set idx 0
	while {1} {
		if {$idx >= [string length $cellpostable]} break
		set vid [get_int cellpostable $idx 2]
		incr idx 3
		set vcid [list $vid [get_int cellpostable $idx 1]]
		incr idx 1
		if {[lsearch $referencedcells $vcid] == -1} {
			lappend referencedcells $vcid
		}
	}
	return $referencedcells
}
proc findAngles {multiangle cellpbtable} {
	if {$multiangle} {return true}
	set idx 0
	while {1} {
		if {$idx >= [string length $cellpbtable]} break
		if {[expr [get_int cellpbtable $idx 1] & 16]==16} {
			return true
		}
		incr idx 24
	}
	return false
}
proc findReferencedCells {type vts} {
	set referencedcells {}
	if {$type == "pgc"} {
		for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
			set referencedcells [addCells $referencedcells $::pgcs($vts,$pgc,cellpostable)]
		}
	} else {
		for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
				set referencedcells [addCells $referencedcells $::menus($vts,$lu,$pgc,cellpostable)]
			}
		}
	}
	return $referencedcells
}

proc find_unref_vob_cells_domain {} {
	if {! [test_version 8.1 "Play all plugin"]} return
	set type [lindex $::editor_cmds_from 0]
	set vts  [lindex [split [lindex $::editor_cmds_from 1] ,] 0]
	set rtype $type
	if {$type == "pgc"} {
		if {$vts == 0} {bell; return}
		set rtype Titel
	}
	set unrefcells [find_unreferenced_cells $type $vts]
	if {$::blocking_msg_aborted} return

	if {[llength $unrefcells] == 0} {
		set txt "Alle Cells von der $rtype Domain des VTS $vts sind referenziert in der IFO."
	} else {
		set txt "Es gibt [llength $unrefcells] Cells in der $rtype Domain\ndes VTS $vts, die nicht in der IFO referenziert sind."
	}
	show_msg "[llength $unrefcells] nicht referenzierte Cells in der $rtype Domain des VTS $vts gefunden."
	tk_messageBox -title "Play All PGC - Plugin" -message $txt
}
proc find_unref_vob_cells_dvd {} {
	if {! [test_version 8.1 "Play all plugin"]} return
	set totalunrefcells 0
	set out "Nicht referenzierte Cells in den VOB-Dateien der DVD:"
	set out "$out\n[file nativename $::dvddir]"
	set out "$out\nDVD-TEXT General Name: \"[::utils::get_General_Name]\""
	set out "$out\nProvider ID: \"[::utils::get_provider_ID]\""
	set out "$out\nAnzahl der VTS: $::pgcs(numvts)"

	set blocking [blocking_msg "Suchen nach Cells in den VOB-Dateien." true]

	for {set vts 0} {$vts <= $::pgcs(numvts)} {incr vts} {
		set out $out\n
		if {$vts == 0} {
			set fname VIDEO_TS.VOB
			if {[file exists [file join $::dvddir $fname]]} {
				set unrefcells [llength [find_unreferenced_cells menu $vts]]
				set out "$out\nVMGM Domain: $unrefcells Cells gefunden in der $fname"
				if {$unrefcells > 0} {set out "$out ***"}
				incr totalunrefcells $unrefcells
			} else {
				set out "$out\nVMGM Domain: Keine $fname Datei."
			}
		} else {
			set fname [format "VTS_%02d_0.VOB" $vts]
			if {[file exists [file join $::dvddir $fname]]} {
				set unrefcells [llength [find_unreferenced_cells menu $vts]]
				set out "$out\nVTSM $vts Domain: $unrefcells Cells gefunden in der $fname"
				if {$unrefcells > 0} {set out "$out ***"}
				incr totalunrefcells $unrefcells
			} else {
				set out "$out\nVTSM $vts Domain: Keine $fname Datei."
			}
		}
		if {$vts > 0} {
			set fname [format "VTS_%02d_1.VOB" $vts]
			if {[file exists [file join $::dvddir $fname]]} {
				set unrefcells [llength [find_unreferenced_cells pgc $vts]]
				set out "$out\nVTST $vts Domain: $unrefcells Cells gefunden in der [format VTS_%02d_*.VOB $vts]"
				if {$unrefcells > 0} {set out "$out ***"}
				incr totalunrefcells $unrefcells
			} else {
				set out "$out\nVTST $vts Domain: Keine $fname Datei! *** FEHLER! ***"
			}
		}
		if {$::blocking_msg_aborted} {
			break
		}
	}

	if {[catch {close_blocking_msg $blocked}]} {close_blocking_msg 1}
	if {$::blocking_msg_aborted} return

	set out "$out\n\n$totalunrefcells nicht referenzierte Cells in der DVD gefunden"
	if {$totalunrefcells > 0} {set out "$out ***"}
	show_msg "$totalunrefcells nicht referenzierte Cells in der DVD gefunden"
	::main::showtext .playallplugin "Play All PGC - Plugin" $out
}
proc find_unreferenced_cells {type vts} {
	set refed [findReferencedCells $type $vts]
	set resultlist {}

	if {$type == "menu"} {
		if {$vts == 0} {
			set vobfile [file join $::dvddir VIDEO_TS.VOB]
		} else {
			set vobfile [file join $::dvddir [format "VTS_%02d_0.VOB" $vts]]
		}
		set vobsize [expr [file size $vobfile] / 2048]
	} else {
		set vobsize 0
		for {set i 1} {$i < 10} {incr i} {
			set vobfile [file join $::dvddir [format VTS_%02d_%d.VOB $vts $i]]
			if {[file exists $vobfile]} {
				incr vobsize [expr [file size $vobfile] / 2048]
			}
		}
		set vobfile [file join $::dvddir [format "VTS_%02d_1.VOB" $vts]]
	}
	if {![file exists $vobfile]} {return {}}

	set blocking [blocking_msg "Suchen nach Cells in den VOB-Dateien." true]

	set err false
	set old_vobfile ""
	set old_vcid {-1 -1}
	set lba 0
	set sector 0
	set fh ""
	set progressbarlimit 0
	while {1} {
		if {$vobfile != $old_vobfile} {
			if {![file exists $vobfile]} {
				set err true
				set errmsg "Datei nicht gefunden"
				break
			}
			if {$fh != ""} {
				close $fh
				set old_vobfile ""
			}
			set fh [::menued::openbinfile $vobfile r]
			if {$fh == ""} {
				set err true
				set errmsg "Kann nicht die [file tail $vobfile] Datei ffnen"
				break
			}
			set old_vobfile $vobfile
		}
		if {$::blocking_msg_aborted} {
			if {[catch {close_blocking_msg $blocked}]} {close_blocking_msg 1}
			show_msg ""
			return {}
		}
		set navpack [::menued::load_nav_pack $vobfile $fh $sector]
		if {$navpack == ""} {
			set err true
			set errmsg "Nav-Pack nicht gltig oder LBA $lba auerhalb des Bereiches"
			break
		}
		set vobuend  [get_int navpack 0x40F 4]
		set vid [get_int navpack 0x41F 2]
		set cid [get_int navpack 0x422 1]
		set vcid [list $vid $cid]
		if {$vcid != $old_vcid && [lsearch resultlist $vcid] == -1} {
			lappend resultlist $vcid
		}
		set progressbarlimit [expr $progressbarlimit + $vobuend + 1]
		if {$progressbarlimit > 25000} {
			show_msg "Scanne $vobfile, VOB/CELL-ID $vid/$cid.  Bitte warten..." $lba $vobsize
			set progressbarlimit 0
		}
		set old_vcid $vcid
		set lba [expr $lba + $vobuend + 1]
		set poslist [::menued::get_nav_pack_position $type $vts $lba $::dvddir]
		set vobfile [lindex $poslist 0]
		set sector  [lindex $poslist 1]
		if {$vobfile == ""} break
		if {$type == "pgc" && ! [file exists $vobfile]} break
		if {[expr [file size $vobfile] / 2048] < $sector} break
	}
	close $fh
	show_msg "Fertig mit dem Scannen." 100
	set unrefed {}
	foreach i $resultlist {
		if {[lsearch $refed $i] == -1} {
			lappend unrefed $i
		}
	}
	if {$blocking && [catch {close_blocking_msg $blocked}]} {close_blocking_msg 1}
	return $unrefed
}

}	;# end of namespace ::plugins::playall
