#############################################################################
# vobu_plugin.tcl Copyright 2006, 2007, 2008, 2009 r0lZ
# PgcEdit plugin to modify the entry and end VOBU sector number of any PGC.
#
# This plugin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This plugin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################
# Usage: copy this file in the <PgcEdit's install dir>/plugins directory.
# A "VOBU" menu will be added in the Plugins menu.
# You need PgcEdit 8.7 or more to use this plugin.
#############################################################################
# Version History:
#
# v 1.6 (December 29, 2008)
# - Little bug: the function "Check for missing VOB or Cell IDs in VOB files"
#   crashed when no DVD was loaded in PgcEdit.
#
# v 1.5 (July 17, 2008)
# - Added the function "Check for missing VOB or Cell IDs in VOB files" to
#   search for gaps in the VOB/Cell IDs numbering sequences.  Although this is
#   not illegal, some apps (like DVD Rebuilder) have problems with the gaps.
#   Use FixVTS or VobBlanker to renumber the VOB/Cell IDs.
#
# v 1.4 (February 16, 2008)
# - Added audio support for the Preview.  Requires PgcEdit v8.7+
#
# v 1.3 (June 20, 2007)
# - Select with Preview was not working any more after the change made in v1.2
# - The help dialog displayed at the beginning can now be hidden.
#
# v 1.2 (June 6, 2007)
# - The temporary file needed to launch the preview is now created with an
#   unique filename to ensure that no conflict exists with another instance
#   of PgcEdit, and is automatically deleted after use.
#
# v 1.1 (September 25, 2006)
# - Added the option to modify also the VTS_C_ADT table
# - Added a way to select the sector number from the preview.
#   (Requires PgcEdit v 7.5 and PgcEdit_Preview v 2.4)
# - Added the "Reset to full range" button to quickly restore the original
#   cell boundaries.
#
# v 1.0 (September 19, 2006)
# - First public release
#############################################################################

if {! [winfo exists .mainmenu.plugins.vobu]} {
	# Add Plugins -> VOBU submenu
	.mainmenu.plugins add cascade -label "VOBU" -menu .mainmenu.plugins.vobu
	menu .mainmenu.plugins.vobu
	.mainmenu.plugins.vobu add command -label "Anfangs- und Endsektor des PGC" -command {::plugins::vobu::pgc_start_end}
	.mainmenu.plugins.vobu add separator
	.mainmenu.plugins.vobu add command -label "berprfen nach fehlenden VOB oder Cell IDs in den VOB-Dateien" -command {::plugins::vobu::check_missing_ids}
	.mainmenu.plugins.vobu add separator
	.mainmenu.plugins.vobu add command -label "ber..." -command {
			tk_messageBox  -title $::plugins::vobu::plugintitle -icon info \
					-message "$::plugins::vobu::plugintitle fr PgcEdit 0.6.0 oder hher.\nVersion $::plugins::vobu::version\nAutor: r0lZ\nFreeware!"
		}
}

namespace eval ::plugins {}
namespace eval ::plugins::vobu {

set version "1.6"
set plugintitle "VOBU Plugin"

#############################################################################
proc pgc_start_end {} {
    if {![test_version 8.7 "Auto Comment Plugin"]} return
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set type  [lindex $::editor_cmds_from 0]
	set index [lindex $::editor_cmds_from 1]
	set vts [lindex [split $index ,] 0]
	if {$type == "menu"} {
		set numcells $::menus($index,numcells)
		if {$numcells == 0} {
			show_msg "Dummy-PGC: Dieses PGC hat kein Video."
			bell
			return
		}
		set cellpbtable $::menus($index,cellpbtable)
		set cellpostable $::menus($index,cellpostable)
		if {$vts == 0} {
			set adtname   VMGM_C_ADT
			set vobutable $::raw_sectors(0,VMGM_VOBU_ADMAP)
		} else {
			set adtname   VTSM_C_ADT
			set vobutable $::raw_sectors($vts,VTSM_VOBU_ADMAP)
		}
	} else {
		set numcells $::pgcs($index,numcells)
		if {$numcells == 0} {
			show_msg "Dummy-PGC: Dieses PGC hat kein Video."
			bell
			return
		}
		set cellpbtable $::pgcs($index,cellpbtable)
		set cellpostable $::pgcs($index,cellpostable)
		set adtname   VTS_C_ADT
		set vobutable $::raw_sectors($vts,VTS_VOBU_ADMAP)
	}
	set adttable  $::raw_sectors($vts,$adtname)

	if {!$::config(expert_mode)} {
		set txt "Diese Funktion erlaubt es den Anfangs-Sektor der VOBU von der ersten"
		set txt "$txt\nCell und den letzten VOBU-Sektor der letzten Cell von dem aktuellen"
		set txt "$txt\nPGC zu ndern."
		set txt "$txt\nBenutzen Sie diese Funktion, um Logos oder Intros zu berspringen."
		set txt "$txt\n"
		set txt "$txt\nWICHTIG: ndern des Anfangs-Sektor oder des letzten VOBU-Sektor ist"
		set txt "$txt\nillegal und kann zu seltsamen Ergebnissen fhren, speziell mit einigen"
		set txt "$txt\nDVD-Playern.  Benutzen dieser Funktion auf eigenes Risiko!"
		set txt "$txt\n"
		set txt "$txt\nAnwendung dieser Funktion:"
		set txt "$txt\nBenutzen Sie die Vorschau, um einen passenden Anfangs oder Endpunkt zu finden."
		set txt "$txt\nNotieren Sie sich die Anzahl, die in der Titelleiste der Vorschau"
		set txt "$txt\n(Absoluter Sektor) angezeigt wird."
    set txt "$txt\nGeben Sie den nchsten VOBU-Sektor in der Liste an. Klicken Sie auf \"Anwenden\"."
		set txt "$txt\nWenn Sie fertig sind, speichern Sie die DVD und Testen Sie es mit Ihrem DVD-Player."
		tk_messageBox -title $::plugins::vobu::plugintitle -icon info -message $txt \
				-parent . -skip vobu_plugin_help
	}

	set origentryvobu1 [get_int cellpbtable 8 4]
	set origfirstilvu1 [get_int cellpbtable 12 4]
	set origlastvobu1  [get_int cellpbtable 16 4]
	set origvid1 [get_int cellpostable 0 2]
	set origcid1 [get_int cellpostable 3 1]

	set cidx [expr ($numcells-1)*24]
	set origentryvobu2 [get_int cellpbtable [expr $cidx+8] 4]
	set origfirstilvu2 [get_int cellpbtable [expr $cidx+12] 4]
	set origlastvobu2  [get_int cellpbtable [expr $cidx+16] 4]
	set numsecs2        [get_int cellpbtable [expr $cidx+4] 3]
	if {$numcells > 1 && ($origentryvobu2 == $origlastvobu2 || $numsecs2 <= 1)} {
		set lastcellistiny true
		incr numcells -1
		set cidx [expr ($numcells-1)*24]
		set next_cell_origentryvobu $origentryvobu2
		set origentryvobu2 [get_int cellpbtable [expr $cidx+8] 4]
		set origfirstilvu2 [get_int cellpbtable [expr $cidx+12] 4]
		set origlastvobu2  [get_int cellpbtable [expr $cidx+16] 4]
		set origcellcmd2   [get_int cellpbtable [expr $cidx+3] 1]
	} else {
		set lastcellistiny false
	}
	set cidx [expr ($numcells-1)*4]
	set origvid2 [get_int cellpostable $cidx 2]
	set origcid2 [get_int cellpostable [expr $cidx+3] 1]

	set idx 8
	while {$idx < [string length $adttable]} {
		set vid [get_int adttable $idx 2]
		incr idx 2
		set cid [get_int adttable $idx 1]
		if {$vid == $origvid1 && $cid == $origcid1} {
			set adtidx1 $idx
			set adtstart1 [get_int adttable [expr $idx+2] 4]
			set adtend1   [get_int adttable [expr $idx+6] 4]
		}
		if {$vid == $origvid2 && $cid == $origcid2} {
			set adtidx2 $idx
			set adtstart2 [get_int adttable [expr $idx+2] 4]
			set adtend2   [get_int adttable [expr $idx+6] 4]
		}
		incr idx 10
	}
	if {$origfirstilvu1 == 0} {
		set adtstart1 [find_start_in_vob $type $vts $adtstart1     $origvid1 $origcid1]
		set tmp       [find_end_in_vob   $type $vts $origlastvobu1 $origvid1 $origcid1]
		if {$tmp != $origlastvobu1} {
			set adtend1 $tmp
		}
	}
	if {$origfirstilvu2 == 0} {
		set adtstart2 [find_start_in_vob $type $vts $adtstart2     $origvid2 $origcid2]
		set tmp       [find_end_in_vob   $type $vts $origlastvobu2 $origvid2 $origcid2]
		if {$tmp != $origlastvobu2} {
			set adtend2 $tmp
		}
	}

	set ::plugins::vobu::validvobus1 {}
	set ::plugins::vobu::validvobus2 {}
	set idx 4
	set tablelen [string length $vobutable]
	while {$idx<$tablelen} {
		set vobu [get_int vobutable $idx 4]
		if {$vobu>=$adtstart1 && $vobu<=$adtend1} {
			lappend ::plugins::vobu::validvobus1 $vobu
		}
		if {$vobu>=$adtstart2 && $vobu<=$adtend2} {
			lappend ::plugins::vobu::validvobus2 $vobu
		}
		incr idx 4
	}

	set w .vobu
	toplevel $w -class Dialog
	wm withdraw $w
	wm title $w "Anfangs-Sektor der VOBU"
	if {$::config(tk) != "aqua"} {wm transient $w .}
	if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
	wm protocol $w WM_DELETE_WINDOW {set ::plugins::vobu::vobu_done 0}
	bind $w <Escape> "set ::plugins::vobu::vobu_done 0"
	bind $w <Return> "set ::plugins::vobu::vobu_done 1"

	label $w.lab -text "Neuer Anfang und letzte VOBUs fr das aktuelle PGC?"
	pack  $w.lab -pady 2

	frame $w.f0 ;# -bd 2 -relief groove
	pack  $w.f0 -side top -fill both -expand 1 -padx 4

	label   $w.f0.lab1 -text "Anfang der VOBU:"
	listbox $w.f0.lb1 -listvariable ::plugins::vobu::validvobus1 \
			-height 24 -width 8 -yscrollcommand "$w.f0.yscr1 set" \
			-selectmode single -activestyle underline -exportselection 0
	set listbox1 $w.f0.lb1
	scrollbar $w.f0.yscr1 -command "$w.f0.lb1 yview"
	if {$::config(use_preview)} {
		if {[test_version 7.5.0 "VOBU Plugin"]} {
			button $w.f0.pvv1 -text "Auswhlen mit Vorschau" -width 20 \
					-command "::plugins::vobu::set_start $type $index 1 $listbox1"
		}
		button $w.f0.pv1 -text "Vorschau der ersten Cell" -width 20 \
				-command "::preview::preview2 $type $index 1 false false 0 false [lindex [::streams::get_best_audio_stream $type $index] 0]"
	}

	label   $w.f0.lab2 -text "Letzte VOBU:"
	listbox $w.f0.lb2 -listvariable ::plugins::vobu::validvobus2 \
			-height 24 -width 8 -yscrollcommand "$w.f0.yscr2 set" \
			-selectmode single -activestyle underline -exportselection 0
	set listbox2 $w.f0.lb2
	scrollbar $w.f0.yscr2 -command "$w.f0.lb2 yview"
	if {$::config(use_preview)} {
		if {[test_version 7.5.0 "VOBU Plugin"]} {
			button $w.f0.pvv2 -text "Auswhlen mit Vorschau" -width 20 \
					-command "::plugins::vobu::set_start $type $index $numcells $listbox2"
		}
		button $w.f0.pv2 -text "Vorschau der letzten Cell" -width 20 \
				-command "::preview::preview2 $type $index $numcells false false 0 false [lindex [::streams::get_best_audio_stream $type $index] 0]"
	}

	frame $w.f0.pad -width 4

	grid $w.f0.lab1		-row 0 -column 0 -sticky w
	grid $w.f0.lb1		-row 1 -column 0 -sticky nsew
	grid $w.f0.yscr1	-row 1 -column 1 -sticky ns
	if {$::config(use_preview)} {
		if {[test_version 7.5.0 "VOBU Plugin"]} {
			grid $w.f0.pvv1	-row 2 -column 0 -sticky ew -columnspan 2
		}
		grid $w.f0.pv1	-row 3 -column 0 -sticky ew -columnspan 2 -pady 2
	}
	grid $w.f0.lab2		-row 0 -column 3 -sticky w
	grid $w.f0.lb2		-row 1 -column 3 -sticky nsew
	grid $w.f0.yscr2	-row 1 -column 4 -sticky ns
	if {$::config(use_preview)} {
		if {[test_version 7.5.0 "VOBU Plugin"]} {
			grid $w.f0.pvv2	-row 2 -column 3 -sticky ew -columnspan 2
		}
		grid $w.f0.pv2	-row 3 -column 3 -sticky ew -columnspan 2 -pady 2
	}
	grid $w.f0.pad		-row 0 -column 2
	grid columnconf $w.f0 0 -weight 1
	grid columnconf $w.f0 3 -weight 1

	if {$::config(use_preview) && [test_version 7.5.0 "VOBU Plugin"]} {
		set txt "Schauen Sie auf das Info-Fenster der Vorschau, um die richtige Sektoradresse herauszufinden"
		set txt "$txt\noder benutzen Sie den Button \"Auswhlen mit Vorschau\" und gehen Sie zu dem Sektor."
		set txt "$txt\nWenn Sie die Vorschau verlassen, wird der aktuell ausgewhlte Sektor verwendet."
		set txt "$txt\nKlicken Sie auf \"Anwenden\", um Ihre nderungen zu bernehmen."
		label $w.lp -text $txt -justify left -anchor w
		pack  $w.lp -padx 4
	}

	if {$lastcellistiny} {
		set txt "HINWEIS: Die letzte Cell (Cell [expr $numcells+1]) ist eine winzige Cell (kleiner als 1 Sekunde)."
		set txt "$txt\nEs ist wahrscheinlich nutzlos, den letzten VOBU-Sektor zu modifizieren."
		set txt "$txt\nDeshalb wird die zweite, anstatt der letzten Cell (Cell $numcells) modifiziert."
		set txt "$txt\nSie werden aufgefordert, dass Seamless Joint Flag von der letzten Cell zu"
		set txt "$txt\nndern, wenn es erforderlich ist."
		label $w.warn -text $txt -justify left -bd 2 -relief groove ;# -bg pink
		pack  $w.warn -padx 4 -fill x
	}

	checkbutton $w.cb -text "Modifizieren auch der $adtname Tabelle?  (Empfohlen)" \
			-variable ::plugins::vobu::modify_vts_c_adt -justify left
	set ::plugins::vobu::modify_vts_c_adt 1
	pack   $w.cb -anchor w

	# buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 4 -fill x
	button $w.fb.rt -text " Zurcksetzen der Auswahl " -command \
		"::plugins::vobu::highlight_item $listbox1 0; ::plugins::vobu::highlight_item $listbox2 end; set ::plugins::vobu::vobu_done 1"
	pack   $w.fb.rt -side left -padx 2
	button $w.fb.cancel -text "Schlieen" -width 10 -command {set ::plugins::vobu::vobu_done 0}
	pack   $w.fb.cancel -side right -padx 2
	button $w.fb.ok -text "Anwenden" -width 10 -command {set ::plugins::vobu::vobu_done 1}
	pack   $w.fb.ok -side right -padx 2

	set txt {{Interleaved or} {multi-angle cell.} {This cell cannot be} {edited.  Sorry!}}
	if {$origfirstilvu1 != 0} {
		set ::plugins::vobu::validvobus1 $txt
		$listbox1 configure -state disabled
	}
	if {$origfirstilvu2 != 0} {
		set ::plugins::vobu::validvobus2 $txt
		$listbox2 configure -state disabled
	}

	update
	wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
	set wx [expr {[winfo rootx .]+(([winfo width  .]-[winfo reqwidth  $w])/2)}]
	set wy [expr {[winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25}]
	if {$wx < 10} {set wx 10}
	if {$wy < 22} {set wy 22}
	wm geometry $w +$wx+$wy

	if {$origfirstilvu1 == 0} {
		set idx [lsearch $::plugins::vobu::validvobus1 $origentryvobu1]
		if {$idx == -1} {set idx 0}
		highlight_item $listbox1 $idx
	}
	if {$origfirstilvu2 == 0} {
		set idx [lsearch $::plugins::vobu::validvobus2 $origlastvobu2]
		if {$idx == -1} {set idx end}
		highlight_item $listbox2 $idx
	}

	wm deiconify $w
	focus $w
	grab set $w

	set done false
	while {!$done} {

		tkwait variable ::plugins::vobu::vobu_done

		if {$::plugins::vobu::vobu_done} {
			if {$origfirstilvu1 == 0 && $numcells==1 && \
					[lindex $::plugins::vobu::validvobus1 [$listbox1 index active]] >
					[lindex $::plugins::vobu::validvobus2 [$listbox2 index active]]} {
				set txt "Da die Anfangs- und End VOBUs von diesem PGC in derselben Cell sind,"
				set txt "$txt\nmuss der End-Sektor grer oder gleich zum Anfangs-Sektor sein."
				set txt "$txt\n\nIhre nderungen werden nicht gespeichert!  Bitte versuchen Sie es erneut."
				tk_messageBox -title $::plugins::vobu::plugintitle -icon error -message $txt -parent .
			} else {
				if {$origfirstilvu1 == 0} {
					set newentryvobu [lindex $::plugins::vobu::validvobus1 [$listbox1 index active]]
					if {$newentryvobu != [get_int cellpbtable 8 4]} {
						set_int cellpbtable 8 $newentryvobu 4
						::main::dvd_modified true true
					}
					if {$::plugins::vobu::modify_vts_c_adt && \
							[get_int adttable [expr $adtidx1+2] 4] != $newentryvobu} {
						set_int adttable [expr $adtidx1+2] $newentryvobu 4
						::main::dvd_modified true true
					}
				}

				if {$origfirstilvu2 == 0} {
					set lidx [$listbox2 index active]
					set newlastvobu [lindex $::plugins::vobu::validvobus2 $lidx]
					if {$lidx < [expr [llength $::plugins::vobu::validvobus2]-1]} {
						set newlastvobuend [expr [lindex $::plugins::vobu::validvobus2 [expr $lidx+1]] - 1]
					} else {
						set newlastvobuend [get_end_of_vobu $type $vts $newlastvobu $vobutable]
					}

					set cidx [expr ($numcells-1)*24]
					if {$newlastvobu != [get_int cellpbtable [expr $cidx+16] 4]} {
						set_int cellpbtable [expr $cidx+16] $newlastvobu 4
						::main::dvd_modified true true
					}
					if {$newlastvobuend != [get_int cellpbtable [expr $cidx+20] 4]} {
						set_int cellpbtable [expr $cidx+20] $newlastvobuend 4
						::main::dvd_modified true true
					}
					if {$::plugins::vobu::modify_vts_c_adt && \
							[get_int adttable [expr $adtidx2+6] 4] != $newlastvobuend} {
						set_int adttable [expr $adtidx2+6] $newlastvobuend  4
						::main::dvd_modified true true
					}

					if {$lastcellistiny} {
						set cidx [expr $numcells*24]
						set flags [get_int cellpbtable $cidx 1]
						# turn off the seamless flag if the last VOBU end is not contiguous
						# to the next cell (the real last cell), and turn it on otherwise.
						if {$next_cell_origentryvobu != [expr $newlastvobuend+1]} {
							set txt "Da die letzte winzige Cell nicht aneinander grenzend ist"
							set txt "$txt\nzum Ende der Cell $numcells, ist es empfehlenswert"
							set txt "$txt\ndas Seamless Joint Flag der letzten Cell auszuschalten."
							set newflags [expr $flags & 0xf7]
							set default yes
						} else {
							set txt "Da die letzte winzige Zelle wieder aneinander grenzend ist"
							set txt "$txt\nzum Ende der Cell $numcells und es kein Cell-Command gibt,"
							set txt "$txt\nfr die Cell $numcells, ist es wahrscheinlich sicher das"
							set txt "$txt\nSeamless Joint Flag von der letzten Cell einzuschalten."
							set txt "$txt\nJedoch ist das nicht zwingend vorgeschrieben."
							if {$origcellcmd2 == 0} {
								set newflags [expr $flags | 8]
							} else {
								set newflags $flags
							}
							set default no
						}
						if {$flags != $newflags} {
							set txt "$txt\n\nModifizieren des Seamless Joint Flags von der letzten Cell?"
							set r [tk_messageBox -title $::plugins::vobu::plugintitle \
									-icon question -message $txt -parent . -type yesno -default $default]
							if {$r == "yes"} {
								set_int cellpbtable $cidx $newflags 1
								::main::dvd_modified true true
							}
						}
					}
				}

				if {$type == "menu"} {
					set ::menus($index,cellpbtable) $cellpbtable
				} else {
					set ::pgcs($index,cellpbtable) $cellpbtable
				}
				if {$::plugins::vobu::modify_vts_c_adt} {
					set ::raw_sectors($vts,$adtname) $adttable
				}
			}
		} else {
			set done true
		}
	}

	destroy $w
	unset -nocomplain ::plugins::vobu::validvobus1
	unset -nocomplain ::plugins::vobu::validvobus2
	unset -nocomplain ::plugins::vobu::lastitem
	return
}

# Find the last sector of the current VOBU
proc get_end_of_vobu {type vts vobustart vobutable} {
	set vobuend $vobustart

	set idx [expr [string length $vobutable] -4]
	set tmp [get_int vobutable $idx 4]
	if {$tmp > $vobustart} {
		set last $tmp
		incr idx -4
		while {$idx>0} {
			set tmp [get_int vobutable $idx 4]
			if {$tmp==$vobustart} {
				set vobuend [expr $last-1]
				break
			}
			set last $tmp
			incr idx -4
		}
	} else {
		set poslist [::menued::get_nav_pack_position $type $vts $vobustart $::dvddir]
		set vobfile [lindex $poslist 0]
		set lba     [lindex $poslist 1]
		set fh [::menued::openbinfile $vobfile r]
		if {$fh != ""} {
			set navpack [::menued::load_nav_pack $vobfile $fh $lba]
			if {$navpack!=""} {
				set vobuend [expr $vobustart+[get_int navpack 0x040F 4]]
			}
		}
	}
	return $vobuend
}

# find the real start of the cell in the VOB file
proc find_start_in_vob {type vts start vid cid} {
	if {$type == "menu"} {
		if {$vts == 0} {
			set filename "VIDEO_TS.VOB"
		} else {
			set filename [format "VTS_%02d_0.VOB" $vts]
		}
	} else {
		set filename [format "VTS_%02d_1.VOB" $vts]
	}
	if {! [file exists [file join $::dvddir $filename]]} {
		return $start
	}
	set poslist  [::menued::get_nav_pack_position $type $vts $start $::dvddir]
	set filename [lindex $poslist 0]
	set sector   [lindex $poslist 1]
	set fh [::menued::openbinfile $filename r]
	if {$fh == ""} {
		return $start
	}
	while {1} {
		set navpack [::menued::load_nav_pack $filename $fh $sector]
		if {$navpack==""} {
			close $fh
			return $start
		}
		if {[get_int navpack 1055 2]!=$vid || [get_int navpack 1058 1]!=$cid} {
			close $fh
			return $start
		}
		for {set a 1425} {$a >= 1349} {incr a -4} {
			set offset [expr [get_int navpack $a 4] & 0xFFFFFFF]
			if {$offset!=268435455 && $offset != 0} {
				set start [expr $start-$offset]
				set poslist  [::menued::get_nav_pack_position $type $vts $start $::dvddir]
				set newfilename [lindex $poslist 0]
				set sector   [lindex $poslist 1]
				if {$newfilename != $filename} {
					close $fh
					set filename $newfilename
					set fh [::menued::openbinfile $filename r]
					if {$fh == ""} {
						return $start
					}
				}
				break
			}
		}
		if {$offset==268435455 || $offset==0} {
			close $fh
			return $start
		}
	}
	close $fh
	return $start
}

# find the real end of the cell in the VOB file
proc find_end_in_vob {type vts end vid cid} {
	if {$type == "menu"} {
		if {$vts == 0} {
			set filename "VIDEO_TS.VOB"
		} else {
			set filename [format "VTS_%02d_0.VOB" $vts]
		}
	} else {
		set filename [format "VTS_%02d_1.VOB" $vts]
	}
	if {! [file exists [file join $::dvddir $filename]]} {
		return $end
	}
	set poslist  [::menued::get_nav_pack_position $type $vts $end $::dvddir]
	set filename [lindex $poslist 0]
	set sector   [lindex $poslist 1]
	set fh [::menued::openbinfile $filename r]
	if {$fh == ""} {
		return $end
	}
	while {1} {
		set navpack [::menued::load_nav_pack $filename $fh $sector]
		if {$navpack==""} {
			close $fh
			return $end
		}
		if {[get_int navpack 1055 2]!=$vid || [get_int navpack 1058 1]!=$cid} {
			close $fh
			return $end
		}
		for {set a 1269} {$a <= 1345} {incr a 4} {
			set offset [expr [get_int navpack $a 4] & 0xFFFFFFF]
			if {$offset!=268435455 && $offset != 0} {
				set end [expr $end+$offset]
				set poslist  [::menued::get_nav_pack_position $type $vts $end $::dvddir]
				set newfilename [lindex $poslist 0]
				set sector   [lindex $poslist 1]
				if {$newfilename != $filename} {
					close $fh
					set filename $newfilename
					set fh [::menued::openbinfile $filename r]
					if {$fh == ""} {
						return $end
					}
				}
				break
			}
		}
		if {$offset==268435455 || $offset==0} {
			close $fh
			return $end
		}
	}
	close $fh
	return $end
}

proc set_start {type index cell listbox} {
	set filename [file normalize [file join $::config(tempdir) "preview_return_[pid].tcl"]]
	catch {file delete $filename}
	::preview::preview2 $type $index $cell false false 0 true [lindex [::streams::get_best_audio_stream $type $index] 0]
	set absolute_lba -1
	catch {source $filename}
	#catch {file delete [file join $::config(tempdir) "preview_args_ini_[pid].tcl"]}
	catch {file delete $filename}
	if {$absolute_lba == -1} {
		bell
		return
	}
	set lst [$listbox get 0 end]
	set idx [lsearch $lst $absolute_lba]
	if {$idx == -1} {
		set idx [lsearch $lst [expr $absolute_lba-1]]
	}
	if {$idx == -1} {
		set mindiff 1000000
		foreach i $lst {
			set diff [expr abs($i-$idx)]
			if {$diff < $mindiff} {
				set mindiff $diff
				set best $i
			} else {
				break
			}
		}
		set idx [lsearch $lst $best]
	}
	highlight_item $listbox $idx
}

proc highlight_item {listbox item} {
	$listbox activate $item
	$listbox selection clear 0 end
	$listbox selection set $item $item
	$listbox see $item
	return
}

proc check_missing_ids {} {
	if {! [info exists ::dvddir]} {bell; return}

	set txt ""
	set scan_needed false
	for {set vts 0} {$vts <= $::pgcs(numvts)} {incr vts} {
		for {set vid 1} {$vid <= $::buttons::butsinfo(menu,$vts,numvobids)} {incr vid} {
		    if {$::buttons::butsinfo(menu,$vts,$vid,numcellids) == 0} {
		        set txt "${txt}Men Domain von VTS $vts:  VOB ID $vid fehlt\n"
		    }
		    for {set cid 1} {$cid <= $::buttons::butsinfo(menu,$vts,$vid,numcellids)} {incr cid} {
		        if {$::buttons::butsinfo(menu,$vts,$vid,$cid,navpack) == -1} {
			        set txt "${txt}Men Domain von VTS $vts:  Cell ID $vid/$cid fehlt\n"
		        }
		    }
		}

		if {$vts == 0} continue
		if {! [info exists ::buttons::butsinfo(pgc,$vts,numvobids)]} {
		    ::menued::get_buttons_all_title_domains $::dvddir true "" true $vts
		}
		if {! [info exists ::buttons::butsinfo(pgc,$vts,numvobids)]} {
		    set txt "${txt}Titel Domain von VTS $vts wurde noch nicht nach BOVs gescannt.\n"
		    set scan_needed true
		    continue
		}
		for {set vid 1} {$vid <= $::buttons::butsinfo(pgc,$vts,numvobids)} {incr vid} {
		    if {$::buttons::butsinfo(pgc,$vts,$vid,numcellids) == 0} {
		        set txt "${txt}Titel Domain von VTS $vts:  VOB ID $vid fehlt\n"
		    }
		    for {set cid 1} {$cid <= $::buttons::butsinfo(pgc,$vts,$vid,numcellids)} {incr cid} {
		        if {$::buttons::butsinfo(pgc,$vts,$vid,$cid,navpack) == -1} {
			        set txt "${txt}Titel Domain von VTS $vts:  Cell ID $vid/$cid fehlt\n"
		        }
		    }
		}
	}
	if {$txt == ""} {
		show_msg "Keine fehlenden VOB oder Cell IDs in der DVD gefunden."
	} else {
		set txt "Fehlende VOB oder Cell IDs in der DVD gefunden:\n\n$txt"
		if {$scan_needed} {
		    set txt "$txt\nSie mssen vor dem Benutzen dieser Funktion alle Titel-Domains nach BOVs scannen\n(selbst wenn sie kein Subpics haben und deshalb keine BOVs haben knnen).\n"
		}
        ::main::showtext .missingvcids "Fehlende VOB oder Cell IDs" $txt
	}
}

}	;# end of namespace ::plugins::vobu
